# -*- coding: utf_8 -*-
from direct.showbase.DirectObject import DirectObject
from pandac.PandaModules import Vec3, BitMask32, Quat
from pandac.PandaModules import OdeTriMeshData, OdeTriMeshGeom
from modules.pcar import playerCar
from car import baseCar
from ai import AI
import pickle

class map(DirectObject):
    def __init__(self,world,rootNode):
        self._world=world
        self.ai=AI(rootNode)
        track=loader.loadModel('res/models/track')
        track.setScale(rootNode,8)
        #self.track.ls()
        #self.track.getTag('scale')
        for i in xrange(5):
            aicar=baseCar(world.world,world.space,Vec3(0,0,15),"AI"+str(i))
            world.cars.append(aicar)
            self.ai.addCar(aicar)
            
            
        self.pcar=playerCar(world.world,world.space,Vec3(0,0,15),"test")  
        world.cars.append(self.pcar)
        sp1=track.find('**/StartPos*')
        self.startCP=self.nearestCheckpoint(sp1.getPos())
        sp=track.findAllMatches('**/StartPos*')
        self.carStates={}
        for i in xrange(sp.getNumPaths()):
            path=sp[i]
            world.cars[i].setCarPosHpr(path.getX()*8,path.getY()*8,path.getZ()*8,path.getH(),path.getP(),path.getR())
            self.carStates[world.cars[i].name]=[0,0,Vec3(),0]
        
        track.flattenLight()
        pborder=track.find('**/Border')
        pborder.copyTo(rootNode)
        pgrass=track.find('**/Grass')
        pgrass.copyTo(rootNode)
        ptrack=track.find('**/Track')
        ptrack.copyTo(rootNode)
        #self.track.reparentTo(rootNode)

        Trimesh = OdeTriMeshData(ptrack, False)
        self.trackGeom = OdeTriMeshGeom(world.space, Trimesh)
        self.trackGeom.setCollideBits(BitMask32(0x00000002))
        self.trackGeom.setCategoryBits(BitMask32.allOff()) 
        
        Trimesh = OdeTriMeshData(pgrass, False)
        self.grassGeom = OdeTriMeshGeom(world.space, Trimesh)
        self.grassGeom.setCollideBits(BitMask32(0x00000002))
        self.grassGeom.setCategoryBits(BitMask32.allOff()) 
        world.space.setSurfaceType(self.grassGeom,4)
        
        Trimesh = OdeTriMeshData(pborder, False)
        self.borderGeom = OdeTriMeshGeom(world.space, Trimesh)
        self.borderGeom.setCollideBits(BitMask32(0x00000002))
        self.borderGeom.setCategoryBits(BitMask32.allOff())
        world.space.setSurfaceType(self.borderGeom,3)
        
        track.removeNode()

        taskMgr.doMethodLater(0.2,self.mapChecker,"mapChecker")

    def mapChecker(self,task):

        for i in xrange(6):
            self.carStates[self._world.cars[i].name][2]=self._world.cars[i].box.getPos()
            if self.nearestCheckpoint(self._world.cars[i].box.getPos())==self.carStates[self._world.cars[i].name][0]:
                self.carStates[self._world.cars[i].name][0]=self.carStates[self._world.cars[i].name][0]+1
                self.carStates[self._world.cars[i].name][3]=self.carStates[self._world.cars[i].name][3]+1
                if self.carStates[self._world.cars[i].name][0]==len(self.ai.path):
                    self.carStates[self._world.cars[i].name][0]=0
                if self.startCP==self.carStates[self._world.cars[i].name][0]:
                    self.carStates[self._world.cars[i].name][1]=self.carStates[self._world.cars[i].name][1]+1            
        infStr=''        
        #infStr=infStr+'\n'+self._world.cars[i].name+':'+str(self.carStates[self._world.cars[i].name][0])+':'+str(self.carStates[self._world.cars[i].name][1])
        #self.carStates.keys().sort()
        for key in self.checkPositions():
            infStr=infStr+key+'\n'
        self.ai.infoText.setText(infStr)
        return task.cont

    def nearestCheckpoint(self,pos):
        n=0
        d=50000
        for i in xrange(len(self.ai.path)):
            x,y,z=self.ai.path[i]
            v=Vec3(pos)-Vec3(x,y,z)
            if v.length()<d:
                d=v.length()
                n=i
        return n
    def checkPositions(self):
        keys=self.carStates.keys()
        sorted=False
        while not sorted:
            sorted=True
            for i in xrange(5):
##                if self.carStates[keys[i]][1]<self.carStates[keys[i+1]][1]:
##                    sorted=False
##                    val=keys[i+1]
##                    keys.remove(val)
##                    keys.insert(i,val)
##                    break
##                if self.carStates[keys[i]][1]==self.carStates[keys[i+1]][1]:
                    if self.carStates[keys[i]][3]<self.carStates[keys[i+1]][3]:
                        sorted=False
                        val=keys[i+1]
                        keys.remove(val)
                        keys.insert(i,val)
                        break
                    if self.carStates[keys[i]][3]==self.carStates[keys[i+1]][3]:
                        p1=self.carStates[keys[i]][2]
                        p2=self.carStates[keys[i+1]][2]
                        cpx,cpy,cpz=self.ai.path[self.carStates[keys[i]][0]]
                        cp=Vec3(cpx,cpy,cpz)
                        vd1=cp-Vec3(p1)
                        vd2=cp-Vec3(p2)
                        if vd1.length()-vd2.length()>10:
                            sorted=False
                            val=keys[i+1]
                            keys.remove(val)
                            keys.insert(i,val)
                            break
        return keys