# -*- coding: utf_8 -*-
from pandac.PandaModules import OdeBody,OdeMass,OdeBoxGeom,OdeSphereGeom #, OdeCylinderGeom
from pandac.PandaModules import OdeHinge2Joint
from pandac.PandaModules import Quat, BitMask32, Vec3, Vec2

#Класс машины/Car class
class playerCar():
    def __init__(self,world,space,pos,name):
        #variables
        #переменные
        self.world=world
        self.world.setContactSurfaceLayer(0.01)
        self.space=space
        self.turn=False
        self.turnspeed=0.0
        self.turnangle=0.0
        self.carOrientation=1
        self.acceleration=False
        self.maxSpeed=0
        self.accForce=0
        self.name=name
        #корпус машины - настравается аналогично кубикам
        #Body of the our car - similar the boxes
        self.box = loader.loadModel("res/models/car")
        self.box.setPos(pos)
        #self.box.setQuat(quat)
        self.box.setColor(1,0.5,0.5)
        self.box.reparentTo(render)     
        self.body=OdeBody(self.world)
        M = OdeMass()
        M.setBox(4, 1.8, 4, 0.5)
        self.body.setMass(M)
        self.body.setPosition(self.box.getPos(render))
        #self.body.setQuaternion(self.box.getQuat(render))
        self.bodyGeom = OdeBoxGeom(self.space, 1,3,1)
        self.bodyGeom.setBody(self.body)
        
        self.joints=[] #подвеска/suspensions
        self.wheelsbody=[] #тело колёс/wheels body
        self.wheelsgeom=[] #геометрия колёс/wheels geometry
        self.wheels=[]     #визуальное представление колёс/wheels visualisation
        for i in range(4):
            #настраиваем физические параметры колеса
            #set physic of the wheel
            self.wheelsbody.append(OdeBody(self.world))
            M = OdeMass()
            M.setCylinder(0.8,2,1, 0.4)
            self.wheelsbody[i].setMass(M)
            self.wheelsbody[i].setQuaternion(Quat(0.7,0,0.7,0))
            self.wheelsbody[i].setFiniteRotationMode(1)
            #self.wheelsgeom.append(OdeCylinderGeom(self.space, 1,0.4))
            self.wheelsgeom.append(OdeSphereGeom(self.space, 1))
            self.wheelsgeom[i].setBody(self.wheelsbody[i])  
            self.wheelsgeom[i].setCategoryBits(BitMask32(0x00000002))
            #добавляем соединение hinge2, имитирующее подвеску
            #add hinge2 joint, wich simulate suspension
            self.joints.append(OdeHinge2Joint(self.world))
            self.joints[i].attachBodies(self.body,self.wheelsbody[i])
            #минимальный и максимальный угол, в пределах которого
            #болтается колесо. У нас оно не должно болтаться, поэтому
            #ставим его одинаковым
            #min/max angle for the wheel. Set min=max for stable turn
            self.joints[i].setParamHiStop(0, 0.0)
            self.joints[i].setParamLoStop(0, 0.0)
            #параметр уменьшения ошибок в подвеске
            #Error reduction parameter of suspension
            self.joints[i].setParamSuspensionERP(0, 0.8)
            #смешивание сил - в данном случае влияет на 
            #жёсткость подвески
            #Blending of forces - in this case influences rigidity of a suspension
            self.joints[i].setParamSuspensionCFM(0, 0.01)            
            #параметр сглаживания рывков при приложении компенсирующих сил
            #self.joints[i].setParamFudgeFactor(0,0.1)
            #оси соединения - одна вертикально, одна горизонтально
            #axis of joint: set one - vertical, and one - horisontal
            self.joints[i].setAxis1(0,0,1)
            self.joints[i].setAxis2(1,0,0)            
            #визуальная модель
            #visual mesh of wheel
            self.wheels.append(loader.loadModelCopy("res/models/wheel"))
            self.wheels[i].setColor(1,0.5,0.5)
            self.wheels[i].setScale(1,1,2)
            self.wheels[i].reparentTo(render)
        
        wheelDistance = 2.4 #1.8
        bodyDistance = 2.2 # 1.1
        #bodyHeight = 2.5
        bodyHeight=0
        #устанавливаем наши колёса в исходое положение
        #set wheels to start position
        self.wheelsbody[0].setPosition(pos.getX()-wheelDistance,pos.getY()+bodyDistance,pos.getZ()+bodyHeight)
        self.wheelsbody[1].setPosition(pos.getX()-wheelDistance,pos.getY()-bodyDistance,pos.getZ()+bodyHeight)
        self.wheelsbody[2].setPosition(pos.getX()+wheelDistance,pos.getY()+bodyDistance,pos.getZ()+bodyHeight)
        self.wheelsbody[3].setPosition(pos.getX()+wheelDistance,pos.getY()-bodyDistance,pos.getZ()+bodyHeight)
        #устанавливаем центы соединений колёс с корпусом
        #set joints to start position
        self.joints[0].setAnchor(Vec3(pos.getX()-(wheelDistance-0.2),pos.getY()+bodyDistance,pos.getZ()+bodyHeight))    
        self.joints[1].setAnchor(Vec3(pos.getX()-(wheelDistance-0.2),pos.getY()-bodyDistance,pos.getZ()+bodyHeight))
        self.joints[2].setAnchor(Vec3(pos.getX()+(wheelDistance-0.2),pos.getY()+bodyDistance,pos.getZ()+bodyHeight))
        self.joints[3].setAnchor(Vec3(pos.getX()+(wheelDistance-0.2),pos.getY()-bodyDistance,pos.getZ()+bodyHeight))
        #Устанавливаем типы поверхности для колёс
        #Set surface types for the wheels
        self.space.setSurfaceType(self.wheelsgeom[0],1)
        self.space.setSurfaceType(self.wheelsgeom[2],1)
        self.space.setSurfaceType(self.wheelsgeom[1],2)
        self.space.setSurfaceType(self.wheelsgeom[3],2)
        
        self.maxVelocity = 65
        self.maxSpeed=50
        self.accForce=500
        
        #регистрируем реакцию на нажатия клавиш и задачу управления машиной
        # register actions and tasks
        axis=[1,3]
        axis2=[1,3,0,2]
        base.accept('w', self.Accel,[self.maxVelocity, 40,axis])
        base.accept('w-up', self.Accel,[0, 15,axis])
        base.accept('s', self.Accel,[-25, 40,axis])
        base.accept('s-up', self.Accel,[0, 15,axis])
        base.accept('space', self.Accel, [0, 200,axis2])
        base.accept('space-up', self.Accel, [0, 15,axis2])
        base.accept('shift', self.Accel, [0, 50,axis2])
        base.accept('shift-up', self.Accel, [0, 15,axis2])
        base.accept('d', self.Turn,[True,0.005])       
        base.accept('a', self.Turn,[True,-0.005])       
        base.accept('d-up', self.Turn,[False,0.005])       
        base.accept('a-up', self.Turn,[False,-0.005])
        taskMgr.add(self.TurnTask,"Rule Car")
        taskMgr.add(self.JetTask,"Jet Task")
        taskMgr.doMethodLater(0.5,self.checkRotation, "checkRotation")
        #Устанавливаем камеру
        #Setup the camera basis
        self.camPosNode = self.box.attachNewNode('camPosNode')
        self.camPosNode.setPos(0,6,-2)
        self.camLookatNode = self.box.attachNewNode('camLookatNode')
        self.camLookatNode.setPos(0,0,2)
        base.camLens.setFar(10000)
        
##        #Спидометр
##        #spedometer
##        spdm = OnscreenImage(image = 'res/textures/spdm.png', scale=0.25, pos = (1, 0, -0.6))
##        spdm.setTransparency(TransparencyAttrib.MAlpha)
##        self.pointer = OnscreenImage(image = 'res/textures/spdm_pointer.png', scale=0.25, pos = (1, 0, -0.6))
##        self.pointer.setTransparency(TransparencyAttrib.MAlpha)
##        self.lastPos = Vec3(0,0,0)
    
    def addCamdist(self, v):
      self.camDistance += v
      print "new camdistance:", self.camDistance
    
    #фнкция придания скорости
    #acceleration function
    def Accel(self, aspect, force, axis):
        for i in [1,3,0,2]:
            self.joints[i].setParamFMax(1, 0)
        #разные методы для движения вперёд и назад
        #вперёд - реактивная тяга; назад - колёса
        #We use two different methods for move forward and backward
        #Forward - "jet engine" - add force to the body of the car
        #Backward - angular engine - add angular speed to the wheels
        if aspect>0:
            self.acceleration=True
        else:
            self.acceleration=False
            for i in axis:
                #устанавливаем скорость вращения
                #set angular engine speed
                self.joints[i].setParamVel(1,aspect*self.carOrientation)
                #и силу с которой мы пытаемся достичь этой скорости
                #and force to it
                self.joints[i].setParamFMax(1, force)
    #проверка ориентации автомобиля и изменение управления в соответствии с этим
    #check car orientation, and change control according to it
    def checkRotation(self,task):
        oldO=self.carOrientation
        if abs(int(self.box.getR()))<90:
            self.carOrientation=1
        else:
            self.carOrientation=-1
        if oldO<>self.carOrientation:
            self.camPosNode.setZ(-self.camPosNode.getZ())
            for i in [1,3,0,2]:
                self.joints[i].setParamVel(1,-self.joints[i].getParamVel(1))
        return task.again
    #поворот колёс (просто присовоение необходимых переменных)
    #turn wheels - set variables
    def Turn(self,enabled,aspect):
        self.turn=enabled
        self.turnspeed=aspect
    #сам поворот осуществляется здесь
    #immediately, turn wheels here
    def TurnTask(self,task):
        #считаем угол
        #calculate angle
        if not self.turn:
            if self.turnangle>0:
                self.turnspeed=-0.01*self.carOrientation
            if self.turnangle<0:
                self.turnspeed=0.01*self.carOrientation
            if -0.01<self.turnangle<0.01:
                self.turnangle=0;
        self.turnangle=self.turnangle+self.turnspeed*self.carOrientation
        if self.turnangle>0.3:
            self.turnangle=0.3
        if self.turnangle<-0.3:
            self.turnangle=-0.3
        # и устанавливаем передним колёсам нужный угол
        # and set angle to the front wheels
        self.joints[0].setParamHiStop(0, self.turnangle)
        self.joints[0].setParamLoStop(0, self.turnangle)
        self.joints[2].setParamHiStop(0, self.turnangle)
        self.joints[2].setParamLoStop(0, self.turnangle)
        # дополнительно фиксируем позицию колёс
        # will fix wheel position a bit better
        for i in xrange(4):
          self.wheelsbody[i].setFiniteRotationAxis(self.joints[i].getAxis2())        
        return task.cont
    #Задача обработки реактивного двигателя
    #task for jet engeene    
    def JetTask(self,task):
        if self.acceleration<>0:
            if self.maxSpeed>self.body.getLinearVel().length():
                self.body.addRelForce(0,self.accForce,0)
        return task.cont
    #синхронизаци видимых частей тележки с физическими
    #sync our visible geometry with them physic bodyes
    def Sync(self):
        self.box.setPos(render, self.body.getPosition())
        self.box.setQuat(render,Quat(self.body.getQuaternion()))
        for i in range(4):
            self.wheels[i].setPos(render, self.wheelsbody[i].getPosition())
            self.wheels[i].setQuat(render,Quat(self.wheelsbody[i].getQuaternion()))
        # обновление положения камеры
        # update the camera
        camVec = self.camPosNode.getPos(render) - self.body.getPosition()
        
        camDistance = Vec2(-4, 2.5)
        targetCamPos = self.body.getPosition() + camVec * camDistance.getX() + Vec3(0,0,camDistance.getY())
        camLookat = self.camLookatNode.getPos(render)
        dPos = targetCamPos - base.camera.getPos(render)
        dt = globalClock.getDt()
        base.camera.setPos(base.camera.getPos(render) + dPos * dt / .5)
        base.camera.lookAt(camLookat)
    
    def setCarPosHpr(self,x,y,z,h,p,r): #установка позиции и поворота
        wheelPos=Vec3(2.2,2.4,0)    #базовая позиция колеса
        l=wheelPos.length()         #длина вектора из базовой позиции
        wheelPos.normalize()
        ad=wheelPos.angleDeg(Vec3(0,1,0))#угол для базовой позиции
        q=Quat()
        q.setHpr(Vec3(h,p,r)) #кватернион из требуемого поворота для корпуса
        self.body.setPosition(Vec3(x,y,z))
        self.body.setQuaternion(q)
        rot=[]
        for i in xrange(4):
            rot.append(Quat())
        #кватернионы для позиции каждого колеса
        rot[0].setHpr(Vec3(h+ad,p,r))
        rot[2].setHpr(Vec3(h-ad,p,r))
        rot[1].setHpr(Vec3(h+(180-ad),p,r))
        rot[3].setHpr(Vec3(h-(180-ad),p,r))

        q.setHpr(Vec3(h,p,r+90)) #кватернион для поворота колеса
        v=[]
        for i in xrange(4):#трансформируем каждый кватернион позиции в вектор и устанавливаем колесо
            v.append(rot[i].getForward()*l)
            self.wheelsbody[i].setPosition(x+v[i].getX(),y+v[i].getY(),z+v[i].getZ())
            self.wheelsbody[i].setQuaternion(q)

##        # стрелка спидометра
##        # the speedometer pointer
##        curPos = self.box.getPos(render)
##        vel = (self.lastPos - curPos).length() * 6000 / self.maxVelocity
##        self.lastPos = curPos
##        dr=vel-self.pointer.getR()
##        if dr>30:
##            dr=30
##        dr=dr*0.1
##        self.pointer.setR(self.pointer.getR()+dr)