# -*- coding: utf_8 -*-
from direct.showbase.DirectObject import DirectObject
from pandac.PandaModules import CollisionTraverser,CollisionNode
from pandac.PandaModules import CollisionHandlerQueue,CollisionRay
from pandac.PandaModules import BitMask32
from random import randint, random
from direct.directtools.DirectGeometry import LineNodePath
from pandac.PandaModules import Vec4,Vec3
import pickle
from pandac.PandaModules import CardMaker


class MousePicker(DirectObject):
    def __init__(self,_render,_camera):
        self.picker         = CollisionTraverser()            #Создание конвейра
        self.pickerQ        = CollisionHandlerQueue()         #Очередь обработки столкновений
        self.pickerCollN    = CollisionNode('mouseRay')       #узел для луча коллизии
        self.pickerCamN     = _camera.attachNewNode(self.pickerCollN) #закрепляем узл на камере
        self.pickerCollN.setFromCollideMask(BitMask32.bit(1))         #устанавливаем маску проверки коллизий
        self.pickerRay = CollisionRay()                #Создаём луч
        self.pickerCollN.addSolid(self.pickerRay)      #и прикрепляем к созданному ранее узлу
        self.picker.addCollider(self.pickerCamN, self.pickerQ) #уведомляем конвейр о новом коллидере
        self.picker.showCollisions(_render)            #визуализируем коллизи (раскомментировать/закоментировать)
        self.accept('mouse1',self.pick)                #регистрируем событие на нажатие кнопки мыши
        self.accept('mouse1-up', self.stopDrag)
        self.accept('control-s',self.savePath)
        self.accept('control-l',self.loadPath,["ai\\test"])
        self.accept('control-x',self.clearPath)
        self._render=_render                           #запоминаем переданные из основной программы
        self._camera=_camera                           #корневой узел и камеру 
        self.pickableObjects=[]                        #создаём список в котором будем хранить список объектов,
                                                       #реагирующих на клик мышью
        self.actionType='startPos'
        self.drag=False
        self.startDragPos=(0,0,0)
        spcm = CardMaker("spcm")
        spcm.setFrame(-10, 10, -20, 20)
        self.posInd = _render.attachNewNode(spcm.generate())
        self.posInd.setPos(0, 0, 5); self.posInd.lookAt(0, 0, -2)
        self.posInd.setTexture(loader.loadTexture('res/textures/start.png'))
        self.pickPoints=[]
        self.pointsToLine=[]
        self.lines = LineNodePath(parent = self._render, thickness = 3.0, colorVec = Vec4(1, 0, 0, 1))
        taskMgr.add(self.drawLines,'drawLines')
        taskMgr.add(self.dragTask,'dragTask') 
    
    #Собственно процедура проверки куда мы ткнули мышой
    def pick(self):
        #Проверяем на наличие мышь
        if base.mouseWatcherNode.hasMouse():
            #если всё ок, то получаем позицию в координатах экрана
            mpos = base.mouseWatcherNode.getMouse()      
            #устанавливаем позицию луча коллизии, основанную на координатах курсора
            self.pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())
            #запускаем конвейр на проверку
            self.picker.traverse(self._render)
            if self.pickerQ.getNumEntries() > 0:
                #если конвейр обнаружил какие-то столкновения, сортируем их
                #что б получить ближайшее
                self.pickerQ.sortEntries()
                i = int(self.pickerQ.getEntry(0).getIntoNode().getTag('obj'))
                #получаем визуальный и физический объекты
                obj=self.pickableObjects[i]
                #получаем точку контакта луча с поверхностью объекта
                p=self.pickerQ.getEntry(0).getSurfacePoint(obj)
                p2=self._render.getRelativePoint(obj, p)
                if self.actionType=='startPos':
                    self.drag=True
                    self.startDragPos=p2
                    self.posInd.setPos(p2)
                    self.posInd.setZ(p2.getZ()+3)
                if self.actionType=='track':
                    self.pickPoints.append((p2.getX(),p2.getY(),p2.getZ()))
                    if len(self.pickPoints)>1:
                        #p3=self.pickPoints[len(self.pickPoints)-2]
                        self.pointsToLine.append(((self.pickPoints[len(self.pickPoints)-2]),
                                                   (self.pickPoints[len(self.pickPoints)-1])))
                        print self.pointsToLine[len(self.pointsToLine)-1]

        
    def dragTask(self,task):
        if self.drag==True:
            if base.mouseWatcherNode.hasMouse():
                mpos = base.mouseWatcherNode.getMouse()      
                self.pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())
                self.picker.traverse(self._render)
                if self.pickerQ.getNumEntries() > 0:
                    self.pickerQ.sortEntries()
                    i = int(self.pickerQ.getEntry(0).getIntoNode().getTag('obj'))
                    obj=self.pickableObjects[i]
                    p=self.pickerQ.getEntry(0).getSurfacePoint(obj)
                    p2=self._render.getRelativePoint(obj, p)
                    p2.setZ(self.startDragPos.getZ())
                    p3=Vec3(p2-self.startDragPos)
                    p3.normalize()
                    ad=p3.angleDeg(Vec3(0,1,0))
                    if p2.getX()<self.startDragPos.getX():
                        ad=-ad
                    self.posInd.setR(ad)
        return task.cont
    def stopDrag(self):
        self.drag=False
    def addPickableObject(self,obj):
        #Устанавливаем такую же маску как и у луча
        obj.find("**/Track").node().setIntoCollideMask(BitMask32.bit(1))
        #Устанавливаем тег, что б потом определить что именно мы пнули
        obj.find("**/Track").node().setTag('obj', str(len(self.pickableObjects)))     
        #втыкаем в список   
        self.pickableObjects.append(obj)
    
    def drawLines(self,task):
        self.lines.reset()
        self.lines.drawLines(self.pointsToLine)
        self.lines.create()
        return task.cont
    
    def savePath(self):
        f=open("ai\\test.path","wb")
        pickle.dump(self.pickPoints,f)
        f.close()
        f=open("ai\\test.pos","wb")
        pos=[(self.posInd.getX(),self.posInd.getY(),self.posInd.getZ()),self.posInd.getR()]
        pickle.dump(pos,f)
        f.close()       
    def loadPath(self,fn):
        f=open(fn+'.path',"r")
        self.pickPoints=[]
        self.pickPoints=pickle.load(f)
        i=0
        self.pointsToLine=[]
        for i in xrange(len(self.pickPoints)):
            if i>0:
                self.pointsToLine.append(((self.pickPoints[i-1]),(self.pickPoints[i])))
            
        f.close()
        f=open(fn+'.pos',"r")
        pos=pickle.load(f)
        px,py,pz=pos[0]
        self.posInd.setPos(px,py,pz)
        self.posInd.setR(pos[1])
        f.close()
    def clearPath(self):
        self.pickPoints=[]
        self.pointsToLine=[]